/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DATAMODEL_H_
#define DATAMODEL_H_

#include "qp_port.h"
#include "dsmapi.h"
#include "selectA-socket.h"

enum DatamodelEnums {
    DATAMODEL_MAGIC                         = 0xDEADBEEF,
    DATAMODEL_STRUCT_VER_MAGIC              = 0xA5A50000,
    DATAMODEL_STRUCT_VER_MAGIC_MASK         = 0xFFFF0000,
    DATAMODEL_STRUCT_VER_MASK               = 0x0000FFFF,
    DATAMODEL_INVALID_ZONE_ID               = 0x0000,
    DATAMODEL_GROUP_ID_NOT_APPLICABLE       = 0xFF,
    DATAMODEL_LOCAL_ZONE_ID_NOT_APPLICABLE  = 0xFF,
    DATAMODEL_DEVICE_STRUCT_VERSION         = DATAMODEL_STRUCT_VER_MAGIC + 7,
    DATAMODEL_GROUP_STRUCT_VERSION          = DATAMODEL_STRUCT_VER_MAGIC + 2,
    DATAMODEL_DSM_INFO_STRUCT_VERSION       = DATAMODEL_STRUCT_VER_MAGIC + 2,
};

enum DatamodelMinMax {
    MIN_GROUP               = 1,
    MAX_GROUP               = 63,
    MAX_GROUPS              = MAX_GROUP-MIN_GROUP+1,
    MIN_BUTTON_GROUP        = 128,
    MIN_ZONE                = 1,
    MAX_ZONE                = 63,
    MAX_ZONES               = MAX_ZONE-MIN_ZONE+1,
    MIN_SCENE               = 0,
    MAX_SCENE               = 63,
    MAX_SCENES              = MAX_SCENE-MIN_SCENE+1,
    MAX_SCENES_1            = 14,
    MAX_SCENES_2            = (MAX_SCENES-MAX_SCENES_1)/2,
    MAX_SCENES_3            = MAX_SCENES-MAX_SCENES_1-MAX_SCENES_2,
    MIN_DEVICE              = 64,
    MAX_DEVICE              = 1022,
    MAX_DEVICES             = MAX_DEVICE-MIN_DEVICE+1,
    MAX_NAME_LEN            = 20,
    MIN_PING_INTERVAL_MS    = 3000,
    MIN_STATE_DIFFERENCES   = 3,
    MIN_APARTMENT_ID        = 0,
    MAX_APARTMENT_ID        = 62,
    MAX_VALID_CIRCUIT_ID    = 62,
    BINARY_INPUT_TABLE_LENGTH = 8,
    SENSOR_TABLE_LENGTH     = 16,
    OPC_TABLE_LENGTH        = 16,
    MAX_PUSH_SENSORS_PER_ZONE = 8,
    HEATING_OPERATION_MODES_LENGTH = 16,
};

enum DatamodelGroups {
    BROADCAST_GROUP     =  0,       /**< broadcast */
    LIGHT_GROUP         =  1,       /**< yellow */
    SHADE_GROUP         =  2,       /**< grey */
    LAST_GROUP_WITH_AREA = SHADE_GROUP,
    NUM_GROUPS_WITH_AREA = LAST_GROUP_WITH_AREA,
    HEATING_GROUP       =  3,       /**< blue */
    AUDIO_GROUP         =  4,       /**< cyan */
    VIDEO_GROUP         =  5,       /**< magenta */
    SECURITY_GROUP      =  6,       /**< red */
    ACCESS_GROUP        =  7,       /**< green */
    JOKER_GROUP         =  8,       /**< black */
    COOLING_GROUP       =  9,
    VENTILATION_GROUP   = 10,
    WINDOW_GROUP        = 11,
    LAST_STANDARD_GROUP = WINDOW_GROUP,
    NUM_STANDARD_GROUPS = LAST_STANDARD_GROUP - 1,              /**< Joker is not handled like standard group */
    USER_GROUP          = 16,                                   /**< first user group */
    FIRST_CLUSTER       = USER_GROUP,
    CLUSTER_1           = 16,
    CLUSTER_2           = 17,
    CLUSTER_3           = 18,
    CLUSTER_4           = 19,
    CLUSTER_5           = 20,
    CLUSTER_6           = 21,
    CLUSTER_7           = 22,
    CLUSTER_8           = 23,
    CLUSTER_24          = 39,
    LAST_CLUSTER        = CLUSTER_24,
    NUM_CLUSTERS        = LAST_CLUSTER - FIRST_CLUSTER + 1,
    MAX_USER_GROUP_WITH_HSM     = LAST_CLUSTER,                 /**< highest user group with HSM support */
    NUM_USER_GROUP_WITH_HSM     = (MAX_USER_GROUP_WITH_HSM - USER_GROUP) + 1,
    FREE_USER_GROUP     = 40,                                   /**< first free user group */
    MIN_USER_GROUP_WITHOUT_HSM  = (MAX_USER_GROUP_WITH_HSM+1),  /**< lowest user group without HSM support */
    ROOM_TEMPERATURE_CONTROL_GROUP  = 48,
    MIN_SPECIAL_FUNCTION_GROUP      = (MAX_GROUP + 1),
    SPECIAL_FUNCTION_SECURITY_GROUP = MIN_SPECIAL_FUNCTION_GROUP,
    SPECIAL_FUNCTION_ACCESS_GROUP,
    MAX_SPECIAL_FUNCTION_GROUP = SPECIAL_FUNCTION_ACCESS_GROUP,
};

/**
 * dSM info flags enumeration
 */
enum DatamodelDsmInfoFlags {
    MMI_SHOW_COLOR_IN_READY_STATE   = 0x01,
    PING_MODE_MASK                  = 0x06,
    ONE_TIME_AUTO_PING_MODE         = 0x02,
    ONE_TIME_OFF_PING_MODE          = 0x04,
    ONE_TIME_OFF_24H_PING_MODE      = 0x06,
    PROGRAMMING_MODE_MASK           = 0x08,
    PROGRAMMING_MODE_DISABLED       = 0x08,
    LOCK_NEW_REGISTERED_DEVICE_MASK = 0x10,
    ROOM_DIMMING_MASK               = 0x80,
    ROOM_DIMMING_DISABLED           = 0x80,
};

enum DatamodelDsmUpstreamFrequencyFlags {
    UPSTREAM_FREQUENCY_PAIR_MASK    = 0x03,
    UPSTREAM_FREQUENCY_15_20        = 0x00,
    UPSTREAM_FREQUENCY_12_20        = 0x01,
    UPSTREAM_FREQUENCY_10_15        = 0x02,
    UPSTREAM_FREQUENCY_10_12        = 0x03,
    UPSTREAM_FREQUENCY_MAX = UPSTREAM_FREQUENCY_10_12,
};

/*..........................................................................*/

/**
 * Data model dSM configuration structure.
 */
typedef struct {
    dsuid_t dsmid;                                  /**< dSM dSUID */
    uint32_t hwRevision;                            /**< hardware revision */
    uint32_t magic;                                 /**< magic */
} DatamodelDsmConfig;

typedef struct {
    uint16_t Kp;                                    /**< range of 1..4000 (resolution if 0.025 deg C) */
    uint16_t Kd;                                    /**< range of 0..1000 */
    uint16_t Ti;                                    /**< range of 1..720 */
    int16_t imin;                                   /**< range of -4000..4000 */
    int16_t imax;                                   /**< range of -4000..4000 */
    uint8_t ymin;                                   /**< range of 100..200 */
    uint8_t ymax;                                   /**< range of 100..200 */
    uint8_t Ts;                                     /**< range of 0..60 */
    uint8_t antiWindUp;                             /**< enable/disable anti wind-up */
    uint16_t sourceZoneId;
    int16_t offset;                                 /**< range of -100..300 */
    uint8_t emergencyValue;                         /**< range of 100..200 */
    uint8_t manualValue;                            /**< range of 100..200 */
    uint8_t keepFloorWarm;
    uint8_t controllerMode;                         /**< mode of the controller */
    uint8_t padding;
} HeatingConfig;

/**
 * Data model dSM info structure
 */
typedef struct {
    uint32_t structId;                              /**< structure version ID */
    uint8_t name[MAX_NAME_LEN];                     /**< dSM name */
    uint8_t apartmentId;                            /**< apartment ID (offset to bus ID on power side) */
    uint8_t warnTemperature;                        /**< warning temperature in degrees */
    uint8_t errorTemperature;                       /**< error temperature in degrees */
    uint8_t flags;                                  /**< flags */
    uint16_t autoPingInterval;                      /**< automatic ping interval time in milliseconds */
    uint16_t oneTimePingInterval;                   /**< one-time ping interval time in milliseconds */
    uint32_t magic;                                 /**< magic */
} DatamodelDsmInfoV1;

typedef struct {
    uint32_t structId;                              /**< structure version ID */
    uint8_t name[MAX_NAME_LEN];                     /**< dSM name */
    uint8_t apartmentId;                            /**< apartment ID (offset to bus ID on power side) */
    uint8_t warnTemperature;                        /**< warning temperature in degrees */
    uint8_t errorTemperature;                       /**< error temperature in degrees */
    uint8_t flags;                                  /**< flags */
    uint8_t upstreamFrequency;                      /**< upstream frequency setting */
    uint16_t autoPingInterval;                      /**< automatic ping interval time in milliseconds */
    uint16_t oneTimePingInterval;                   /**< one-time ping interval time in milliseconds */
    uint32_t magic;                                 /**< magic */
} DatamodelDsmInfoV2;

typedef DatamodelDsmInfoV2 DatamodelDsmInfo;

#define DATAMODEL_DSM_INFO_MAX_LEN                    max(sizeof(DatamodelDsmInfoV1), sizeof(DatamodelDsmInfoV2))

/**
 * Data model max temperature structure
 */
typedef struct {
    uint32_t structId;                              /**< structure version ID */
    int16_t maxTemperature;                         /**< max measured temperature */
    uint32_t magic;                                 /**< magic */
} DatamodelMaxTemperature;

/**
 * Data model device structure
 */
typedef struct {
    uint32_t structId;                              /**< structure version ID */
    uint32_t dsid;                                  /**< device serial number (global ID) */
    uint16_t deviceId;                              /**< device address (local ID) */
    uint16_t vid;                                   /**< device vendor ID */
    uint16_t fid;                                   /**< device function ID */
    uint16_t pid;                                   /**< device product ID */
    uint16_t ver;                                   /**< device version */
    uint8_t locked;                                 /**< device entry locked */
    uint8_t localZoneId;                            /**< local zone ID the device is in */
    uint8_t myClick;                                /**< my click number 0: standard button, 1..4: myClick button */
    uint8_t mode;                                   /**< cached device MODE register */
    uint8_t groupMembership[8];                     /**< group membership of the device (bit-field) 1: is member 0: is not member */
    uint8_t buttonGroup[16];                        /**< assigns each of the 16 dSD-buttons to a group */
    uint8_t buttonZoneMembership[2];                /**< button zone membership (bit-field) 1: button is in dSD's zone, 0: button is in all zones */
    uint8_t name[MAX_NAME_LEN];                     /**< device name */
    uint32_t magic;                                 /**< magic */
} DatamodelDeviceV1;

typedef struct {
    uint32_t structId;                              /**< structure version ID */
    uint32_t dsid;                                  /**< device serial number (global ID) */
    uint16_t deviceId;                              /**< device address (local ID) */
    uint16_t vid;                                   /**< device vendor ID */
    uint16_t fid;                                   /**< device function ID */
    uint16_t pid;                                   /**< device product ID */
    uint16_t ver;                                   /**< device version */
    uint8_t locked;                                 /**< device entry locked */
    uint8_t localZoneId;                            /**< local zone ID the device is in */
    uint8_t groupMembership[8];                     /**< group membership of the device (bit-field) */
    uint8_t mode;                                   /**< cached device MODE register */
    uint8_t buttonIdAndGroup;                       /**< cached button ID (button behavior) and color (group) of push-button */
    uint8_t buttonUserGroup;                        /**< overridden user group of push-button or 0xFF (not applicable) */
    bool buttonSetLocalPriority;                    /**< set if local priority is set after group scene call */
    uint8_t name[MAX_NAME_LEN];                     /**< device name */
    uint32_t magic;                                 /**< magic */
} DatamodelDeviceV2;

typedef struct {
    uint32_t structId;                              /**< structure version ID */
    char dsid[25];                                  /**< device dSID (global ID) */
    uint16_t deviceId;                              /**< device address (local ID) */
    uint16_t vid;                                   /**< device vendor ID */
    uint16_t fid;                                   /**< device function ID */
    uint16_t pid;                                   /**< device product ID */
    uint16_t ver;                                   /**< device version */
    uint8_t locked;                                 /**< device entry locked */
    uint8_t localZoneId;                            /**< local zone ID the device is in */
    uint8_t groupMembership[8];                     /**< group membership of the device (bit-field) */
    uint8_t ltmode;                                 /**< cached device LTMODE register */
    uint8_t mode;                                   /**< cached device MODE register */
    uint8_t buttonIdAndGroup;                       /**< cached button ID (button behavior) and color (group) of push-button */
    uint8_t buttonUserGroup;                        /**< overridden user group of push-button or 0xFF (not applicable) */
    bool buttonSetLocalPriority;                    /**< set if local priority is set after group scene call */
    uint8_t name[MAX_NAME_LEN];                     /**< device name */
    uint32_t magic;                                 /**< magic */
    bool active;
} DatamodelDeviceV3;

typedef struct {
    uint8_t targetGroup;
    uint8_t type;
    uint8_t NUM;
    uint8_t present;
} BinaryInputConfig;

typedef struct {
    union {
        struct {
            uint8_t pollInterval : 3;
            uint8_t broadcast    : 1;
            uint8_t pushConvert  : 1;
            uint8_t reserved     : 3;
        };
        uint8_t config;
    };
    uint8_t type;
} SensorConfig;

typedef struct {
    uint8_t channelId;
} OpcConfig;

typedef struct {
    uint32_t structId;                              /**< structure version ID */
    uint32_t dsid;                                  /**< device serial number (global ID) */
    uint16_t deviceId;                              /**< device address (local ID) */
    uint16_t vid;                                   /**< device vendor ID */
    uint16_t fid;                                   /**< device function ID */
    uint16_t pid;                                   /**< device product ID */
    uint16_t ver;                                   /**< device version */
    uint8_t locked;                                 /**< device entry locked */
    uint8_t localZoneId;                            /**< local zone ID the device is in */
    uint8_t groupMembership[8];                     /**< group membership of the device (bit-field) */
    uint8_t ltmode;                                 /**< cached device LTMODE register */
    uint8_t mode;                                   /**< cached device MODE register */
    uint8_t buttonIdAndGroup;                       /**< cached button ID (button behavior) and color (group) of push-button */
    uint8_t buttonUserGroup;                        /**< overridden user group of push-button or 0xFF (not applicable) */
    bool buttonSetLocalPriority;                    /**< set if local priority is set after group scene call */
    uint8_t name[MAX_NAME_LEN];                     /**< device name */
    BinaryInputConfig binaryInputTable[BINARY_INPUT_TABLE_LENGTH]; /**< a copy of the binary input table of the device */
    uint32_t magic;                                 /**< magic */
} DatamodelDeviceV4;

typedef struct {
    uint32_t structId;                              /**< structure version ID */
    char dsid[25];                                  /**< device dSID (global ID) */
    uint16_t deviceId;                              /**< device address (local ID) */
    uint16_t vid;                                   /**< device vendor ID */
    uint16_t fid;                                   /**< device function ID */
    uint16_t pid;                                   /**< device product ID */
    uint16_t ver;                                   /**< device version */
    uint8_t locked;                                 /**< device entry locked */
    uint8_t localZoneId;                            /**< local zone ID the device is in */
    uint8_t groupMembership[8];                     /**< group membership of the device (bit-field) */
    uint8_t ltmode;                                 /**< cached device LTMODE register */
    uint8_t mode;                                   /**< cached device MODE register */
    uint8_t buttonIdAndGroup;                       /**< cached button ID (button behavior) and color (group) of push-button */
    uint8_t buttonUserGroup;                        /**< overridden user group of push-button or 0xFF (not applicable) */
    union {
        struct {
            uint8_t buttonSetLocalPriority: 1;          /**< set if local priority is set after group scene call */
            uint8_t buttonNoComingHomeCall: 1;          /**< set if a button click doesn't trigger a coming home call when apartment is in absent mode */
            uint8_t buttonApartment:        1;          /**< set if the button acts on the whole apartment (zone 0) */
            uint8_t reserved:               5;          /**< reserved for later usage */
        };
        uint8_t flags;
    };
    uint8_t name[MAX_NAME_LEN];                     /**< device name */
    BinaryInputConfig binaryInputTable[BINARY_INPUT_TABLE_LENGTH]; /**< a copy of the binary input table of the device */
    uint32_t magic;                                 /**< magic */
    bool active;
} DatamodelDeviceV5;

typedef struct {
    uint32_t structId;                              /**< structure version ID */
    char dsid[25];                                  /**< device dSID (global ID) */
    uint16_t deviceId;                              /**< device address (local ID) */
    uint16_t vid;                                   /**< device vendor ID */
    uint16_t fid;                                   /**< device function ID */
    uint16_t pid;                                   /**< device product ID */
    uint16_t ver;                                   /**< device version */
    uint8_t locked;                                 /**< device entry locked */
    uint8_t localZoneId;                            /**< local zone ID the device is in */
    uint8_t groupMembership[8];                     /**< group membership of the device (bit-field) */
    uint8_t ltmode;                                 /**< cached device LTMODE register */
    uint8_t mode;                                   /**< cached device MODE register */
    uint8_t buttonIdAndGroup;                       /**< cached button ID (button behavior) and color (group) of push-button */
    uint8_t buttonUserGroup;                        /**< overridden user group of push-button or 0xFF (not applicable) */
    union {
        struct {
            uint8_t buttonSetLocalPriority: 1;          /**< set if local priority is set after group scene call */
            uint8_t buttonNoComingHomeCall: 1;          /**< set if a button click doesn't trigger a coming home call when apartment is in absent mode */
            uint8_t buttonApartment:        1;          /**< set if the button acts on the whole apartment (zone 0) */
            uint8_t reserved:               5;          /**< reserved for later usage */
        };
        uint8_t flags;
    };
    uint8_t name[MAX_NAME_LEN];                     /**< device name */
    BinaryInputConfig binaryInputTable[BINARY_INPUT_TABLE_LENGTH]; /**< a copy of the binary input table of the device */
    SensorConfig sensorTable[SENSOR_TABLE_LENGTH];  /**< a copy of the sensor table of the device */
    uint8_t padding[3];
    uint32_t magic;                                 /**< magic */
    bool active;
} DatamodelDeviceV6;

typedef struct {
    uint32_t structId;                              /**< structure version ID */
    char dsuid[DSUID_STR_LEN];                      /**< device dSUID (global ID) */
    uint16_t deviceId;                              /**< device address (local ID) */
    uint16_t vid;                                   /**< device vendor ID */
    uint16_t fid;                                   /**< device function ID */
    uint16_t ver;                                   /**< device version */
    uint8_t locked;                                 /**< device entry locked */
    uint8_t localZoneId;                            /**< local zone ID the device is in */
    uint8_t groupMembership[8];                     /**< group membership of the device (bit-field) */
    uint8_t ltmode;                                 /**< cached device LTMODE register */
    uint8_t mode;                                   /**< cached device MODE register */
    uint8_t buttonIdAndGroup;                       /**< cached button ID (button behavior) and color (group) of push-button */
    uint8_t buttonUserGroup;                        /**< overridden user group of push-button or 0xFF (not applicable) */
    uint8_t buttonOutputChannel;                    /**< cached device PBCHANNEL register */
    uint8_t buttonGlobalGroup;                      /**< cached device PBGROUP register */
    union {
        struct {
            uint8_t buttonSetLocalPriority: 1;          /**< set if local priority is set after group scene call */
            uint8_t buttonNoComingHomeCall: 1;          /**< set if a button click doesn't trigger a coming home call when apartment is in absent mode */
            uint8_t buttonApartment:        1;          /**< set if the button acts on the whole apartment (zone 0) */
            uint8_t reserved:               5;          /**< reserved for later usage */
        };
        uint8_t flags;
    };
    uint8_t name[MAX_NAME_LEN];                     /**< device name */
    BinaryInputConfig binaryInputTable[BINARY_INPUT_TABLE_LENGTH]; /**< a copy of the binary input table of the device */
    SensorConfig sensorTable[SENSOR_TABLE_LENGTH];  /**< a copy of the sensor table of the device */
    OpcConfig opcTable[OPC_TABLE_LENGTH];           /**< a copy of the opc table of the device */
    uint8_t padding;
    uint32_t magic;                                 /**< magic */
    bool active;
    char vdcdsuid[DSUID_STR_LEN];                   /**< vdc dSUID (global ID) */
} DatamodelDeviceV7;

typedef DatamodelDeviceV7 DatamodelDevice;

#define DATAMODEL_DEVICE_MAX_LEN                    max(max(max(max(max(max( \
                                                        sizeof(DatamodelDeviceV1), \
                                                        sizeof(DatamodelDeviceV2)), \
                                                        sizeof(DatamodelDeviceV3)), \
                                                        sizeof(DatamodelDeviceV4)), \
                                                        sizeof(DatamodelDeviceV5)), \
                                                        sizeof(DatamodelDeviceV6)), \
                                                        sizeof(DatamodelDeviceV7))

/**
 * Data model group structure part 1
 * */
typedef struct {
    uint8_t id;                                     /**< group ID */
    uint8_t stateMachineGroupId;                    /**< group ID of state machine */
    uint8_t name[MAX_NAME_LEN];                     /**< group name */
    uint8_t sceneName[MAX_SCENES_1][MAX_NAME_LEN];  /**< scene names of a group */
    uint32_t magic;                                 /**< magic */
} DatamodelGroup1V1;

typedef struct {
    uint32_t structId;                              /**< structure version ID */
    uint8_t id;                                     /**< group ID */
    uint8_t stateMachineGroupId;                    /**< group ID of state machine */
    uint8_t stopTimeout;                            /**< STOP timeout for this group */
    uint8_t remoteStopTimeout;                      /**< remotly set STOP timeout for this group */
    uint8_t name[MAX_NAME_LEN];                     /**< group name */
    uint8_t sceneName[MAX_SCENES_1][MAX_NAME_LEN];  /**< scene names of a group */
    uint32_t magic;                                 /**< magic */
} DatamodelGroup1V2;

typedef DatamodelGroup1V2 DatamodelGroup1;

/**
 * Data model group structure part 2
 * */
typedef struct {
    uint8_t sceneName[MAX_SCENES_2][MAX_NAME_LEN];  /**< scene names of a group */
    uint32_t magic;                                 /**< magic */
} DatamodelGroup2;

/**
 * Data model group structure part 3
 * */
typedef struct {
    uint8_t sceneName[MAX_SCENES_3][MAX_NAME_LEN];  /**< scene names of a group */
    uint32_t magic;                                 /**< magic */
} DatamodelGroup3;

/**
 * Data model cluster structure
 * */
typedef struct {
    uint8_t id;                                     /**< cluster ID */
    uint8_t stateMachineGroupId;                    /**< cluster ID of state machine */
    uint8_t stopTimeout;                            /**< STOP timeout for this cluster */
    uint8_t remoteStopTimeout;                      /**< remotely set STOP timeout for this cluster */
    uint8_t name[MAX_NAME_LEN];                     /**< cluster name */
    uint16_t location;                              /**< location of the cluster */
    uint16_t floor;                                 /**< floor of the cluster */
    uint16_t protectionClass;                       /**< protection class of the cluster */
    bool configurationLock;                         /**< 0=full configuration allowed; 1=cluster is locked */
    uint8_t sceneLock[16];                          /**< bit-field of locked scenes */
} DatamodelCluster;

/* Data model zone structure */
typedef struct {
    uint16_t zoneId;                                /**< zone ID (global ID) */
    uint8_t localZoneId;                            /**< zone ID (local ID) */
    uint8_t zoneTagId;                              /**< zone tag ID used to detect newer zone */
    uint8_t name[MAX_NAME_LEN];                     /**< zone name */
    uint32_t magic;                                 /**< magic */
} DatamodelZone;

/* Data model device state structure */
typedef struct {
    uint32_t structId;                              /**< structure version ID */
    uint8_t tagId;                                  /**< active devices tag ID to detect newest active devices structure */
    uint8_t activeDevices[(MAX_DEVICE+7)/8];        /**< active devices bit-field */
    uint32_t magic;                                 /**< magic */
} DatamodelDeviceState;

/*..........................................................................*/

/* Data model zone info structure */
typedef struct {
    DatamodelZone super;
    uint8_t groupMembership[(MAX_GROUP+7)/8];
} DatamodelZoneInfo;

/*..........................................................................*/

/**
 * Data model publish configuration event
 */
typedef struct {
    QEvent super;                   /**< base event structure */
    dsuid_t dsmid;                  /**< published dSM ID */
    uint16_t autoPingInterval;      /**< automatic ping interval time in milliseconds */
    uint16_t oneTimePingInterval;   /**< one-time ping interval time in milliseconds */
    uint8_t apartmentId;            /**< published apartment ID */
    uint8_t dsmInfoFlags;           /**< published dSM info flags */
} DatamodelConfigEvt;

/**
 * Data model publish zone id event
 */
typedef struct {
    QEvent super;
    uint16_t zoneId;            /**< published global zone ID */
    uint8_t localZoneId;        /**< local zone ID of published global zone ID */
} DatamodelZoneIdEvt;

/**
 * Data model basic request event
 */
typedef struct {
    QEvent super;
    QActive * requester;        /**< requesting active object pointer */
} DatamodelReqBasicEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    dsuid_t dsuid;
    uint8_t clusterId;
} DatamodelReqClusterInfoEvt;

/**
 * Data model dSM request event
 */
typedef struct {
    QEvent super;
    dsuid_t dsuid;
    QActive * requester;        /**< requesting active object pointer */
} DatamodelReqDsmEvt;

/**
 * Data model set dSM configuration request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    dsuid_t dsmid;
    uint32_t hwRevision;
} DatamodelReqSetDsmConfig;

/**
 * Data model zone modify request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
    uint16_t newZoneId;
    uint8_t modifier;
    dsuid_t dsuid;
} DatamodelReqZoneModifyEvt;

/**
 * Data model zone device count request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    ZoneInfoModifiers_t zoneIdModifier;
    uint8_t modifier;
    uint16_t zoneId;
    dsuid_t dsuid;
} DatamodelReqZoneDeviceCountEvt;

/**
 * Data model device info of index request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    ZoneInfoModifiers_t zoneIdModifier;
    uint16_t zoneId;
    uint16_t deviceIndex;
    uint8_t modifier;
    uint8_t privateId;
    dsuid_t dsuid;
} DatamodelReqDeviceInfoOfIndexEvt;

/**
 * Data model zone info of index request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t zoneIndex;
    dsuid_t dsuid;
} DatamodelReqZoneInfoOfIndexEvt;

/**
 * Data model zone info by id request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
} DatamodelReqZoneInfoByIdEvt;

/**
 * Data model zone local id request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
} DatamodelReqZoneLocalIdEvt;

/**
 * Data model device request event (given the device ID)
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceId;
    uint8_t privateId;
} DatamodelReqDeviceIdEvt;

/**
 * Data model device request event (given the device serial number)
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    char dsuid[DSUID_STR_LEN];
    uint8_t privateId;
} DatamodelReqDeviceSerialEvt;

/**
 * Data model group membership modify request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t groupId;
    uint8_t modifier;
    bool sendDownstream;
} DatamodelReqGroupMembershipModifyEvt;

/**
 * Data model get name request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t primaryId;         /* primary ID: device ID or zone ID */
    uint16_t secondaryId;       /* secondary ID: group ID or scene/group ID */
    uint8_t modifier;
} DatamodelReqGetNameEvt;

/**
 * Data model set name request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t primaryId;         /* primary ID: device ID or zone ID */
    uint16_t secondaryId;       /* secondary ID: group ID or scene/group ID */
    uint8_t modifier;
    uint8_t name[MAX_NAME_LEN];
} DatamodelReqSetNameEvt;

/**
 * Data model add device request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    char dsuid[DSUID_STR_LEN];      /**< device dSUID (global ID) */
    uint16_t vid;                   /**< device vendor ID */
    uint16_t fid;                   /**< device function ID */
    uint16_t ver;                   /**< device firmware version */
    char vdcdsuid[DSUID_STR_LEN];   /**< device dSUID (global ID) */
    uint16_t zoneId;                /**< zoneID of device or 0, if not available */
} DatamodelReqAddDeviceEvt;

/**
 * Data model ping device state (active/inactive) request event
 */
typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceId;                      /**< device ID */
    uint16_t oneTimePingInitialDeviceId;    /**< one-time ping initial device ID (0 = invalid) */
    bool active;                            /**< device active (TRUE) / inactive (FALSE) flag */
} DatamodelReqPingDeviceStateEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
    uint8_t modifier;
    uint8_t groupId;
    dsuid_t vdcdsuid;
} DatamodelReqZoneGroupInfoEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
    uint8_t groupId;
    uint8_t stateMachineGroupId;
} DatamodelReqZoneGroupAddEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
    uint8_t groupId;
    dsuid_t dsuid;
} DatamodelReqZoneGroupRemoveEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
} DatamodelReqDeviceGroupsEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
} DatamodelReqDeviceInfoEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint16_t zoneId;
} DatamodelReqDeviceSetZoneEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t buttonNumber;
} DatamodelReqDeviceGetButtonGroupEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t groupId;
} DatamodelReqDeviceSetButtonGroupEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    bool setLocalPriority;
} DatamodelReqDeviceSetButtonLocalPriorityEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    bool noComingHome;
} DatamodelReqDeviceSetButtonNoComingHomeEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    bool apartment;
} DatamodelReqDeviceSetButtonApartmentEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t channelId;
    bool sendDownstream;
} DatamodelReqDeviceSetButtonOutputChannelEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t group;
    bool sendDownstream;
} DatamodelReqDeviceSetButtonGlobalGroupEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t MODERegister;
} DatamodelReqDeviceSetMODERegEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t LTMODERegister;
} DatamodelReqDeviceSetLTMODERegEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t buttonIdAndGroupRegister;
} DatamodelReqDeviceSetButtonIdAndGroupRegEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    BinaryInputConfig table[BINARY_INPUT_TABLE_LENGTH];
} DatamodelReqDeviceSetBinaryInputTableEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t offset;
    uint8_t value;
} DatamodelReqDeviceSetBinaryInputTableByteEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    SensorConfig table[SENSOR_TABLE_LENGTH];
} DatamodelReqDeviceSetSensorTableEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    OpcConfig table[OPC_TABLE_LENGTH];
} DatamodelReqDeviceSetOpcTableEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t offset;
    uint8_t value;
} DatamodelReqDeviceSetSensorTableByteEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    bool locked;
} DatamodelReqDeviceSetLockedEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t name[MAX_NAME_LEN];
    dsuid_t dsuid;
} DatamodelReqNameEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t apartmentId;
    dsuid_t dsuid;
} DatamodelReqSetApartmentIdEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t warnTemperature;
    uint8_t errorTemperature;
    dsuid_t dsuid;
} DatamodelReqSetTemperatureLimitsEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t autoPingInterval;                  /**< automatic ping interval time in milliseconds */
    uint16_t oneTimePingInterval;               /**< one-time ping interval time in milliseconds */
    dsuid_t dsuid;
} DatamodelReqSetPingIntervalsEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t flags;
    dsuid_t dsuid;
} DatamodelReqSetDsmFlagsEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t upstreamFrequency;
    dsuid_t dsuid;
} DatamodelReqSetDsmUpstreamFreqEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
    uint8_t name[MAX_NAME_LEN];
    dsuid_t dsuid;
} DatamodelReqZoneNameEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
    uint8_t groupId;
    uint8_t name[MAX_NAME_LEN];
} DatamodelReqGroupNameEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
    uint8_t groupId;
    uint8_t stopTimeout;
} DatamodelReqGroupStopTimeoutEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
    uint8_t groupId;
    uint8_t color;
} DatamodelReqGroupColorEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t clusterId;
    uint8_t name[MAX_NAME_LEN];
} DatamodelReqClusterNameEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t clusterId;
    uint8_t stopTimeout;
} DatamodelReqClusterStopTimeoutEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t clusterId;
    uint8_t color;
} DatamodelReqClusterColorEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t clusterId;
    uint16_t location;
    uint16_t floor;
    uint16_t protectionClass;
} DatamodelReqClusterLocationClassEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t clusterId;
    uint8_t sceneLock[16];
} DatamodelReqClusterSceneLockEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t clusterId;
    bool lock;
} DatamodelReqClusterConfigLockEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint8_t localZoneId;
    uint8_t groupId;
} DatamodelReqUserGroupInfoEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
    uint8_t groupId;
    uint8_t sceneIndex;
    uint8_t name[MAX_NAME_LEN];
} DatamodelReqSceneNameEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
    uint8_t name[MAX_NAME_LEN];
} DatamodelReqDeviceNameEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t zoneId;
    uint8_t groupId;
    uint8_t sceneIndex;
} DatamodelReqSceneInfoEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint32_t dataPacketIndex;
} DatamodelReqFlashReadEvt;

typedef struct {
    DatamodelReqBasicEvt basic;
    uint16_t deviceAddress;
} DatamodelReqRemoveDeviceEvt;

typedef struct {
    QEvent super;
    uint16_t deviceId;
    uint16_t sensorValue;
    uint8_t sensorIndex;
} DatamodelReqDeviceSensorPushEvt;

/*..........................................................................*/

typedef struct {
    QEvent super;
    int16_t maxTemperature;                     /**< max temperature */
    int16_t currentTemperature;                 /**< current temperature */
    uint16_t autoPingInterval;                  /**< automatic ping interval time in milliseconds */
    uint16_t oneTimePingInterval;               /**< one-time ping interval time in milliseconds */
    uint8_t apartmentId;                        /**< apartment ID */
    uint8_t warnTemperature;                    /**< warn temperature limit */
    uint8_t errorTemperature;                   /**< error temperature limit */
    uint8_t flags;                              /**< flags */
    uint8_t upstreamFrequency;                  /**< Upstream frequency setting */
} DatamodelResDsmProperties;

/**
 * Data model error response event
 */
typedef struct {
    QEvent super;
    uint8_t error;
} DatamodelErrorEvt;

/**
 * Data model long error response event
 */
typedef struct {
    QEvent super;
    uint8_t error;
    uint8_t privateId;
} DatamodelLongErrorEvt;

/**
 * Data model device entry response event
 */
typedef struct {
    QEvent super;               /**< base event */
    DatamodelDevice device;     /**< device structure */
    bool state;                 /**< active: TRUE, inactive: FALSE */
    uint8_t privateId;          /**< private ID supplied by */
} DatamodelResDeviceEntryEvt;

typedef struct {
    QEvent super;
    bool active;
} DatamodelResDeviceStateEvt;

/**
 * Data model zone count response event
 */
typedef struct {
    QEvent super;
    uint8_t zoneCount;
} DatamodelResZoneCountEvt;

/**
 * Data model zone device count response event
 */
typedef struct {
    QEvent super;
    uint16_t deviceCount;
} DatamodelResZoneDeviceCountEvt;

/**
 * Data model zone info response event
 */
typedef struct {
    QEvent super;
    uint16_t zoneId;
    uint8_t localZoneId;
    uint8_t groupCount;
    uint8_t name[MAX_NAME_LEN];
} DatamodelResZoneInfoEvt;

/**
 * Data model zone local id response event
 */
typedef struct {
    QEvent super;
    uint8_t localZoneId;
} DatamodelResZoneLocalIdEvt;

/**
 * Data model add device response event
 */
typedef struct {
    QEvent super;
    uint16_t deviceId;              /**< device ID of added device */
    uint8_t localZoneId;            /**< local zone ID where added device has been put */
    uint16_t zoneId;                /**< zone ID where added device has been put */
} DatamodelResAddDeviceEvt;

/**
 * Data model ping device active/inactive response event
 */
typedef struct {
    QEvent super;
    uint32_t nextDeviceSerial;      /**< device serial of next device ID to ping */
    uint16_t nextDeviceId;          /**< next device ID to ping */
    bool oneTimePingDone;           /**< one-time ping is done */
} DatamodelResPingDeviceEvt;

typedef struct {
    QEvent super;
    uint8_t groupId;
    uint8_t stateMachineGroupId;
    uint16_t countDevices;
    uint8_t name[MAX_NAME_LEN];
    uint8_t stopTimeout;
    uint8_t remoteStopTimeout;
} DatamodelResZoneGroupInfoEvt;

typedef struct {
    QEvent super;
    uint16_t groupId;
} DatamodelResZoneGroupModifyEvt;

typedef struct {
    QEvent super;
    uint8_t groupMembership[8];
} DatamodelResDeviceGroupsEvt;

typedef struct {
    QEvent super;
    uint8_t groupId;
} DatamodelResGetMyActionGroupEvt;

typedef struct {
    QEvent super;
    dsuid_t dsuid;
} DatamodelResDsuidEvt;

typedef struct {
    QEvent super;
    uint8_t clusterId;
    uint8_t stateMachineGroupId;
    uint16_t countDevices;
    uint8_t name[MAX_NAME_LEN];
    uint8_t stopTimeout;
    uint8_t remoteStopTimeout;
    uint8_t configLock;
    uint8_t sceneLock[16];
    uint8_t padding;
    uint16_t location;
    uint16_t floor;
    uint16_t protectionClass;
} DatamodelResClusterInfoEvt;

typedef struct {
    QEvent super;
    uint32_t hwVersion;
    uint32_t armSwVersion;
    dsuid_t dsuid;
    uint8_t name[MAX_NAME_LEN];
} DatamodelResDsmInfoEvt;

typedef struct {
    QEvent super;
    uint8_t name[MAX_NAME_LEN];
} DatamodelResSceneInfoEvt;

typedef struct {
    QEvent super;
    uint16_t P_W;
    int16_t P_Q;
} DatamodelResGetCurrentMeterValueEvt;

typedef struct {
    QEvent super;
    uint8_t localZoneId;
    uint8_t groupId;
    uint8_t stateMachineGroupId;
} DatamodelResUserGroupInfoEvt;

typedef struct {
    QEvent super;
    uint32_t hash;
    uint32_t changeCount;
} DatamodelResHashEvt;

typedef struct {
    QEvent super;
    uint8_t localZoneId;
    uint8_t groupId;
} DatamodelReqStopTimeoutUpdateEvt;

/*..........................................................................*/

/**
 * Data model constructor
 */
void Datamodel_ctor(dsuid_t id, const char* databasePath);

/**
 * Opaque pointer to data model AO
 */
extern QActive * const AO_Datamodel;

uint8_t Datamodel_DeviceButtonZone(DatamodelDevice const * device);
uint8_t Datamodel_DeviceButtonActiveGroup(DatamodelDevice const * device);
uint8_t Datamodel_DeviceButtonId(DatamodelDevice const * device);
uint8_t Datamodel_DeviceButtonGroup(DatamodelDevice const * device);
uint8_t Datamodel_ButtonId_ButtonGroupId(uint8_t buttonId);
DatamodelZoneInfo * Datamodel_GetZoneInfoByLocalId(uint8_t localId);
bool Datamodel_DsmPropertiesProgrammingModeDisabled();
bool Datamodel_IsDeviceValid(uint16_t deviceId);
bool Datamodel_GetDsuidFromDeviceId(uint16_t deviceId, char *dsuid);
uint16_t Datamodel_GetDeviceIdFromDsuid(const char *dsuid);
uint8_t Datamodel_GetDeviceLtmode(uint16_t deviceId);
void Datamodel_SetDeviceSensor(uint16_t deviceId, uint8_t index, uint8_t type, uint32_t pollInterval, uint8_t broadcast, uint8_t pushConvert);
void Datamodel_SetDeviceSensorName(uint16_t deviceId, uint8_t index, const char *name);
void Datamodel_RemoveSpareSensorEntries(uint16_t deviceId, uint8_t index);
uint8_t Datamodel_GetDeviceSensorCount(uint16_t deviceId);
bool Datamodel_GetDeviceSensor(uint16_t deviceId, uint8_t index, uint8_t *type, uint32_t *pollInterval, uint8_t *broadcast, uint8_t *pushConvert);
bool Datamodel_GetDeviceSensorType(uint16_t deviceId, uint8_t index, uint8_t *type);
void Datamodel_SetDeviceBinaryInput(uint16_t deviceId, uint8_t index, uint8_t group, uint8_t type, uint8_t buttonId);
void Datamodel_SetDeviceBinaryInputName(uint16_t deviceId, uint8_t index, const char *name);
uint8_t Datamodel_GetDeviceBinaryInputCount(uint16_t deviceId);
void Datamodel_RemoveSpareBinaryInputs(uint16_t deviceId, uint8_t index);
bool Datamodel_GetDeviceBinaryInput(uint16_t deviceId, uint8_t index, uint8_t *targetGroup, uint8_t *type, uint8_t *buttonId);
bool Datamodel_GetDeviceBinaryInputType(uint16_t deviceId, uint8_t index, uint8_t *type);
void Datamodel_SetDeviceGroupMembershipAddGroup(uint16_t deviceId, uint8_t groupId);
void Datamodel_ClearDeviceGroupMembership(uint16_t deviceId);
void Datamodel_DeriveGroupMembership(uint16_t deviceId);

void Datamodel_updateDeviceVdc(uint16_t deviceId, char *vdcdsuid, uint16_t fid);


uint8_t Datamodel_updateDeviceZone( /*I/O parameter */ uint16_t* zoneId, uint16_t deviceId, const char* vdcdsuid_str);

bool Datamodel_DsmPropertiesRoomDimmingDisabled();
uint8_t Datamodel_GetStopTimeoutZoneGroup(uint8_t localZoneId, uint8_t groupId);
dsuid_t Datamodel_GetDsmid();
bool Datamodel_IsCluster(uint8_t groupId);

void Datamodel_ZoneGroupGetDeviceDsuids(uint8_t localZoneId, uint8_t groupId, device_dsuid_t **dsuid_list);

void Datamodel_SetOpcTableValue(uint16_t deviceId, uint8_t index, uint8_t channelId);
uint8_t Datamodel_GetOpcChannelCount(uint16_t deviceId);

bool Datamodel_GetOpcChannelFromIndex(uint16_t deviceId, uint8_t index, uint8_t *channelId);
void Datamodel_RemoveSpareOpcTableEntries(uint16_t deviceId, uint8_t index);

uint8_t calculateCircuitId(uint8_t busAddress, uint8_t apartmentId);

void Datamodel_add_vdc(const char *vdcDsuid);

bool Datamodel_GetVdcDsuidFromDeviceId(uint16_t deviceId, dsuid_t *dsuid);

bool Datamodel_GetVdcDsuidFromDeviceDsuid(dsuid_t *deviceDsuid, dsuid_t *vdcDsuid);

uint8_t Datamodel_GetBusMemberType(const dsuid_t *vdcDsuid);

uint8_t Datamodel_SetSensorPushDevice(uint8_t localZoneId, uint8_t sensorType, dsuid_t *sensorDsuid);

uint8_t Datamodel_GetSensorPushDevice(uint8_t localZoneId, uint8_t sensorType, dsuid_t *sensorDsuid);

uint8_t Datamodel_ResetSensorPushDevice(uint8_t localZoneId, uint8_t sensorType);


/* for external access: vdc awareness */
bool Datamodel_GetDsuidFromHeatingZone(uint16_t zoneId, dsuid_t *dsuid);

uint8_t Datamodel_setHeatingConfigOfvDC(const dsuid_t *vdcDsuid, uint16_t zoneId, HeatingConfig *heating);

uint8_t Datamodel_getHeatingConfigOfvDC(const dsuid_t *vdcDsuid, uint16_t zoneId, HeatingConfig *heating);

uint8_t Datamodel_getHeatingOperationModesOfvDC(const dsuid_t *vdcDsuid, uint16_t zoneId, uint16_t *operationMode, uint16_t length);

/* for internal access */
uint8_t Datamodel_setHeatingConfig(dsuid_t const *vdcDsuid, uint16_t zoneId, HeatingConfig *heating);

uint8_t Datamodel_getHeatingConfig(uint16_t zoneId, HeatingConfig *heating);

uint8_t Datamodel_setHeatingOperationModes(uint16_t zoneId, uint16_t *operationMode, uint16_t length);

uint8_t Datamodel_getHeatingOperationModes(uint16_t zoneId, uint16_t *operationMode, uint16_t length);

uint8_t Datamodel_updateVdcName(const char *vdc_dsuid, const char *name);

uint8_t Datamodel_getCluster(uint8_t clusterId, DatamodelCluster* cluster);


void Datamodel_setVdcDefaultZoneID(const char *vdcDsuid, uint16_t defaultZoneId);
uint16_t Datamodel_getVdcDefaultZoneID(const char *vdcDsuid);
uint16_t Datamodel_create_zoneId_from_dsuid(dsuid_t const *dsuid);
uint64_t Datamodel_groups_array_to_int(uint8_t *groups);
uint64_t Datamodel_sceneLock_array_to_int(const uint8_t *sceneLock, uint8_t sceneLockSize);

#endif /* DATAMODEL_H_ */
