/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <functional>
#include <boost/noncopyable.hpp>
#include <avahi-client/client.h>
#include <avahi-client/publish.h>
#include "poll.h"

namespace ds {
namespace avahi {

class EntryGroup;

/// Client connection to avahi deamon
class Client : boost::noncopyable {
public:
    Client(Poll& poll);

    ::AvahiClient* get() const { return m_ptr.get(); }
    boost::asio::io_service& getIoService() { return m_poll.getIoService(); }

    /// Add and publish entry group.
    void addEntryGroup(EntryGroup &);

    // throw error stored in `Client` object by avahi stack.
    void throwError(const char* message);
    void throwError(const char* message, int error);

    void handleError(const std::exception &e);

private:
    Poll& m_poll;
    struct Deleter {
        void operator()(::AvahiClient *x) { avahi_client_free(x); }
    };
    std::unique_ptr< ::AvahiClient, Deleter> m_ptr;
    bool m_active;
    std::vector<EntryGroup*> m_entryGroups;

    void setActive(bool active);
    void reconnect();

    static void sCallback(AvahiClient *, AvahiClientState state, void *data);
    void callback(AvahiClientState state);
};

} // namespace avahi
} // namespace ds
