/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <functional>
#include <boost/noncopyable.hpp>
#include <avahi-client/client.h>
#include <avahi-client/publish.h>
#include "poll.h"

namespace ds {
namespace avahi {

class Client;

struct Service {
  AvahiIfIndex interface;
  AvahiProtocol protocol;
  AvahiPublishFlags flags;
  std::string name;
  std::string type;
  std::string domain;
  std::string host;
  uint16_t port;
  std::vector<std::string> txts;

  explicit Service() : interface(AVAHI_IF_UNSPEC), protocol(AVAHI_PROTO_UNSPEC), flags(), port() {}
  Service(Service &&) = default;
};

enum struct CollisionMode {
    ALTERNATIVE_NAME, ///< Retry with alternative name on collision
    //TODO(someday): RETRY,  ///< Retry with the same name on collision
};

class EntryGroup {
public:
    EntryGroup(Client& client, CollisionMode collisioMode, std::vector<Service>&& services = std::vector<Service>());

    ::AvahiEntryGroup* get() const { return m_ptr.get(); }

    /// Get raw services data. Call \ref publish() to activate any changes.
    /// Returned reference is invalidated on any non-const call.
    std::vector<Service>& getServices() { return m_services; }
    const std::vector<Service>& getServices() const { return m_services; }

    /// Publish changes done to services data.
    void publish();

private:
    Client& m_client;
    struct Deleter {
        void operator()(::AvahiEntryGroup *x) { avahi_entry_group_reset(x); }
    };
    std::unique_ptr< ::AvahiEntryGroup, Deleter> m_ptr;
    CollisionMode m_collisionMode;
    std::vector<Service> m_services;

    bool getActive() { return static_cast<bool>(m_ptr); }
    void setActive(bool active);

    void handleCollision();

    static void sCallback(AvahiEntryGroup *g, AvahiEntryGroupState state, void *data);
    void callback(AvahiEntryGroupState state);

    friend class Client;
};

} // namespace avahi
} // namespace ds
