/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "client.h"
#include <avahi-common/error.h>
#include "util.h"

typedef boost::system::error_code error_code;

namespace ds {
namespace avahi {

Poll::Poll(boost::asio::io_service &ioService) : m_ioService(ioService), m_timer(ioService) {
    VDSM_ENTER();
    m_simplePoll.reset(avahi_simple_poll_new());
    if (!m_simplePoll) {
        throw std::runtime_error("Failed to create simple poll object.\n");
    }

    asyncIterate();
}

// Poor man's avahi loop integration into asio
void Poll::asyncIterate()
{
    m_timer.expires_from_now(boost::chrono::milliseconds(200));
    m_timer.async_wait([this](const error_code& e)
    {
        if (e) {
            return; //timer was aborted
        }

        // Function avahi_simple_poll_iterate will not tell us whether it has more
        // events to process or not.
        // Processing only 1 event per timeout leads to slow startup.
        // TODO(someday): write proper avahi loop integration
        for (int i = 0; i < 8; i++) {
            if (avahi_simple_poll_iterate(m_simplePoll.get(), 0) != 0) {
                throw std::runtime_error("Avahi poll iterate error");
            }
        }
        asyncIterate();
    });
}

} // namespace avahi
} // namespace ds
