/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <boost/noncopyable.hpp>
#include <avahi-common/simple-watch.h>
#include <ds/asio/timer.h>


namespace ds {
namespace avahi {

/// Avahi event poll integration to asio
class Poll : boost::noncopyable {
public:
    Poll(boost::asio::io_service &ioService);

    const AvahiPoll* get() const { return avahi_simple_poll_get(m_simplePoll.get()); }
    boost::asio::io_service& getIoService() { return m_ioService; }

private:
    boost::asio::io_service& m_ioService;
    ds::asio::Timer m_timer;

    struct SimplePollDeleter {
        void operator()(::AvahiSimplePoll *x) { avahi_simple_poll_free(x); }
    };
    std::unique_ptr< ::AvahiSimplePoll, SimplePollDeleter> m_simplePoll;

    void asyncIterate();
};

} // namespace avahi
} // namespace ds
