/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <stdexcept>
#include "str.h"

namespace ds {

// * `DS_ASSUME(condition)`:  Check preconditions -- e.g. to validate parameters passed from a caller.
// A failure indicates that the caller is buggy.
#define DS_ASSUME(cond) \
  if (DS_LIKELY(cond)) {} else \
    throw std::runtime_error(ds::str(__FILE__, ':', __LINE__, ' ', #cond))

// * `DS_CHECK(condition)`:  Check condition, throw if the condition does not hold.
#define DS_CHECK(cond) \
  if (DS_LIKELY(cond)) {} else \
    throw std::runtime_error(ds::str(__FILE__, ':', __LINE__, ' ', #cond))

} // namespace ds

